'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var vueVirtualTextdocument = require('@vuedx/vue-virtual-textdocument');
var FS = _interopDefault(require('fs'));
var Path = _interopDefault(require('path'));
var vscodeUri = require('vscode-uri');

function isNumber(any) {
    return typeof any === 'number';
}
function flat(items) {
    const result = [];
    items.forEach(item => {
        if (Array.isArray(item)) {
            result.push(...flat(item));
        }
        else {
            result.push(item);
        }
    });
    return result;
}
function firstOrDefault(items, defaultValue) {
    if (items.length)
        return items[0];
    return defaultValue();
}
function chain(value) {
    return {
        next(fn) {
            if (value == undefined)
                return chain(value);
            return chain(fn(value));
        },
        // @ts-ignore
        end(defaultValue) {
            return value == undefined ? defaultValue : value;
        },
    };
}
class VueContext {
    constructor(documents) {
        this.documents = documents;
    }
    getFileNameFromVirtualFileName(fileName) {
        return fileName.split(vueVirtualTextdocument.virtualFileNameSep).shift();
    }
    isVueFile(fileName) {
        return vueVirtualTextdocument.isVueFile(fileName);
    }
    isVirtualFile(fileName) {
        return vueVirtualTextdocument.isVirtualFile(fileName);
    }
    getVueDocument(fileNameOrUri) {
        return this.documents.get(vueVirtualTextdocument.asUri(fileNameOrUri));
    }
    getVirtualDocument(fileNameOrUri) {
        return chain(vueVirtualTextdocument.parseVirtualFileUri(fileNameOrUri))
            .next(ref => { var _a; return ((_a = this.getVueDocument(ref.uri)) === null || _a === void 0 ? void 0 : _a.getBlockDocument(ref.selector)) || null; })
            .end(null);
    }
    getVirtualDocumentAt(fileName, position) {
        const document = this.getVueDocument(fileName);
        if (!document)
            return null;
        return document.getBlockDocument(document.blockAt(position)) || null;
    }
    getFileNameAt(fileName, position) {
        var _a;
        if (this.isVueFile(fileName)) {
            return (_a = this.getVirtualDocumentAt(fileName, position)) === null || _a === void 0 ? void 0 : _a.fsPath;
        }
        return fileName;
    }
    getInnerFileNames(fileName, onlyScript = false) {
        if (this.isVueFile(fileName)) {
            const document = this.documents.get(fileName);
            if (!document)
                return [];
            return [document.getBlockDocument('script').fsPath];
        }
        return [fileName];
    }
}
class VueLanguageServer {
    constructor(context, service, typescript) {
        this.context = context;
        this.service = service;
        this.typescript = typescript;
    }
    cleanupSemanticCache() {
        throw new Error('Method not implemented.');
    }
    getSyntacticDiagnostics(fileName) {
        return flat(this.context
            .getInnerFileNames(fileName)
            .map(this.service.getSyntacticDiagnostics));
    }
    getSemanticDiagnostics(fileName) {
        return flat(this.context
            .getInnerFileNames(fileName)
            .map(this.service.getSemanticDiagnostics));
    }
    getSuggestionDiagnostics(fileName) {
        return flat(this.context
            .getInnerFileNames(fileName)
            .map(this.service.getSuggestionDiagnostics));
    }
    getCompilerOptionsDiagnostics() {
        return this.service.getCompilerOptionsDiagnostics();
    }
    getSyntacticClassifications(fileName, span) {
        return flat(this.context
            .getInnerFileNames(fileName, true)
            .map(fileName => this.service.getSyntacticClassifications(fileName, span)));
    }
    getSemanticClassifications(fileName, span) {
        return flat(this.context
            .getInnerFileNames(fileName, true)
            .map(fileName => this.service.getSemanticClassifications(fileName, span)));
    }
    getEncodedSyntacticClassifications(fileName, span) {
        return firstOrDefault(this.context
            .getInnerFileNames(fileName, true)
            .map(fileName => this.service.getEncodedSyntacticClassifications(fileName, span)), () => ({ spans: [], endOfLineState: this.typescript.EndOfLineState.None }));
    }
    getEncodedSemanticClassifications(fileName, span) {
        return firstOrDefault(this.context
            .getInnerFileNames(fileName, true)
            .map(fileName => this.service.getEncodedSyntacticClassifications(fileName, span)), () => ({ spans: [], endOfLineState: this.typescript.EndOfLineState.None }));
    }
    getCompletionsAtPosition(fileName, position, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getCompletionsAtPosition(fileName, position, options))
            .end();
    }
    getCompletionEntryDetails(fileName, position, name, formatOptions, source, preferences) {
        throw chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getCompletionEntryDetails(fileName, position, name, formatOptions, source, preferences))
            .end();
    }
    getCompletionEntrySymbol(fileName, position, name, source) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.getCompletionEntrySymbol(fileName, position, name, source))
            .end();
    }
    getQuickInfoAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getQuickInfoAtPosition(fileName, position))
            .end();
    }
    getNameOrDottedNameSpan(fileName, startPos, endPos) {
        return chain(this.context.getFileNameAt(fileName, startPos))
            .next(fileName => this.service.getNameOrDottedNameSpan(fileName, startPos, endPos))
            .end();
    }
    getBreakpointStatementAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getBreakpointStatementAtPosition(fileName, position))
            .end();
    }
    getSignatureHelpItems(fileName, position, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getSignatureHelpItems(fileName, position, options))
            .end();
    }
    getRenameInfo(fileName, position, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getRenameInfo(fileName, position, options))
            .next(info => {
            if ('fileToRename' in info && info.fileToRename) {
                info.fileToRename = this.context.getFileNameFromVirtualFileName(info.fileToRename);
            }
            return info;
        })
            .end({ canRename: false, localizedErrorMessage: '' });
    }
    findRenameLocations(fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.findRenameLocations(fileName, position, findInStrings, findInComments, providePrefixAndSuffixTextForRename))
            .next(locations => locations.map(location => {
            console.log('vue:: remaping => ' + location.fileName);
            location.fileName = this.context.getFileNameFromVirtualFileName(location.fileName);
            return location;
        }))
            .end();
    }
    getSmartSelectionRange(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getSmartSelectionRange(fileName, position))
            .end({ textSpan: { start: position, length: 0 } });
    }
    getDefinitionAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getDefinitionAtPosition(fileName, position))
            .end();
    }
    getDefinitionAndBoundSpan(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getDefinitionAndBoundSpan(fileName, position))
            .end();
    }
    getTypeDefinitionAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getTypeDefinitionAtPosition(fileName, position))
            .end();
    }
    getImplementationAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getImplementationAtPosition(fileName, position))
            .end();
    }
    getReferencesAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getReferencesAtPosition(fileName, position))
            .end();
    }
    findReferences(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.findReferences(fileName, position))
            .end();
    }
    getDocumentHighlights(fileName, position, filesToSearch) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getDocumentHighlights(fileName, position, filesToSearch))
            .end();
    }
    getOccurrencesAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getOccurrencesAtPosition(fileName, position))
            .end();
    }
    getNavigateToItems(searchValue, maxResultCount, fileName, excludeDtsFiles) {
        return this.service.getNavigateToItems(searchValue, maxResultCount, fileName, excludeDtsFiles);
    }
    getNavigationBarItems(fileName) {
        return flat(this.context
            .getInnerFileNames(fileName)
            .map(fileName => this.service.getNavigationBarItems(fileName)));
    }
    getNavigationTree(fileName) {
        return firstOrDefault(this.context
            .getInnerFileNames(fileName)
            .map(fileName => this.service.getNavigationTree(fileName)), () => ({
            kind: this.typescript.ScriptElementKind.moduleElement,
            text: '<unknown>',
            kindModifiers: 'none',
            nameSpan: undefined,
            spans: [],
        }));
    }
    prepareCallHierarchy(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.prepareCallHierarchy(fileName, position))
            .end();
    }
    provideCallHierarchyIncomingCalls(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.provideCallHierarchyIncomingCalls(fileName, position))
            .end([]);
    }
    provideCallHierarchyOutgoingCalls(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.provideCallHierarchyOutgoingCalls(fileName, position))
            .end([]);
    }
    getOutliningSpans(fileName) {
        return flat(this.context
            .getInnerFileNames(fileName, true)
            .map(fileName => this.service.getOutliningSpans(fileName)));
    }
    getTodoComments(fileName, descriptors) {
        return flat(this.context
            .getInnerFileNames(fileName, true)
            .map(fileName => this.service.getTodoComments(fileName, descriptors)));
    }
    getBraceMatchingAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getBraceMatchingAtPosition(fileName, position))
            .end([]);
    }
    getIndentationAtPosition(fileName, position, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getIndentationAtPosition(fileName, position, options))
            .end(0);
    }
    getFormattingEditsForRange(fileName, start, end, options) {
        return chain(this.context.getFileNameAt(fileName, start))
            .next(fileName => this.service.getFormattingEditsForRange(fileName, start, end, options))
            .end([]);
    }
    getFormattingEditsForDocument(fileName, options) {
        return flat(this.context
            .getInnerFileNames(fileName, true)
            .map(fileName => this.service.getFormattingEditsForDocument(fileName, options)));
    }
    getFormattingEditsAfterKeystroke(fileName, position, key, options) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getFormattingEditsAfterKeystroke(fileName, position, key, options))
            .end([]);
    }
    getDocCommentTemplateAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getDocCommentTemplateAtPosition(fileName, position))
            .end();
    }
    isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.isValidBraceCompletionAtPosition(fileName, position, openingBrace))
            .end(false);
    }
    getJsxClosingTagAtPosition(fileName, position) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getJsxClosingTagAtPosition(fileName, position))
            .end();
    }
    getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
        return chain(this.context.getFileNameAt(fileName, position))
            .next(fileName => this.service.getSpanOfEnclosingComment(fileName, position, onlyMultiLine))
            .end();
    }
    getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
        return chain(this.context.getFileNameAt(fileName, start))
            .next(fileName => this.service.getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences))
            .end([]);
    }
    getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
        return this.service.getCombinedCodeFix(scope, fixId, formatOptions, preferences);
    }
    applyCodeActionCommand(action, formatSettings) {
        return this.service.applyCodeActionCommand(action, formatSettings);
    }
    getApplicableRefactors(fileName, positionOrRange, preferences) {
        return chain(this.context.getFileNameAt(fileName, isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos))
            .next(fileName => this.service.getApplicableRefactors(fileName, positionOrRange, preferences))
            .end([]);
    }
    getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
        return chain(this.context.getFileNameAt(fileName, isNumber(positionOrRange) ? positionOrRange : positionOrRange.pos))
            .next(fileName => this.service.getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences))
            .end();
    }
    organizeImports(scope, formatOptions, preferences) {
        return this.service.organizeImports(scope, formatOptions, preferences);
    }
    getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
        return this.service.getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences);
    }
    getEmitOutput(fileName, emitOnlyDtsFiles, forceDtsEmit) {
        return this.service.getEmitOutput(fileName, emitOnlyDtsFiles, forceDtsEmit);
    }
    getProgram() {
        return this.service.getProgram();
    }
    dispose() {
        return this.service.dispose();
    }
    toLineColumnOffset(fileName, position) {
        return this.service.toLineColumnOffset(fileName, position);
    }
    static create(context, server, typescript) {
        const instance = new VueLanguageServer(context, server, typescript);
        const proxy = Object.create(null);
        const keys = Object.keys(server);
        for (const key of keys) {
            proxy[key] = instance[key].bind(instance);
        }
        return proxy;
    }
}

function init({ typescript: ts }) {
    console.log('Init vue plugin.');
    function create(info) {
        info.project.projectService.logger.info('vue:: Create ts server with vue plugin...');
        function log(...args) {
            info.project.projectService.logger.info('vue:: ' + args.join(' '));
        }
        console.log = log;
        const context = new VueContext(new vueVirtualTextdocument.DocumentStore(uri => {
            info.project.projectService.logger.info('vue:: store load ' + uri);
            return vueVirtualTextdocument.VueTextDocument.create(uri, 'vue', 0, FS.readFileSync(vscodeUri.URI.parse(uri).fsPath, { encoding: 'utf8' }) || '');
        }, () => info.project.useCaseSensitiveFileNames()));
        function patchFileSystem(key) {
            if (info[key].fileExists) {
                override(info[key], 'fileExists', fn => path => vueVirtualTextdocument.isVirtualFile(path) || fn(path));
            }
            if (info[key].readFile) {
                override(info[key], 'readFile', fn => (path) => vueVirtualTextdocument.isVirtualFile(path)
                    ? context.getVirtualDocument(path).getText()
                    : fn(path));
            }
            if (info[key].readDirectory) {
                override(info[key], 'readDirectory', fn => (path, extensions, exclude, include, depth) => {
                    if (extensions)
                        extensions = ['vue'].concat(extensions);
                    return fn(path, extensions, include, exclude, depth);
                });
            }
        }
        function patchModuleResolution(key) {
            const cache = ts.createModuleResolutionCache(info.project.getCurrentDirectory(), fileName => fileName);
            override(info[key], 'resolveModuleNames', fn => function (moduleNames, containingFile, reusedNames, redirectedReference, options) {
                if (containingFile.includes('node_modules')) {
                    // TODO: Maybe handle in node_modules
                    return fn(moduleNames, containingFile, reusedNames, redirectedReference, options);
                }
                if (vueVirtualTextdocument.isVirtualFile(containingFile)) {
                    containingFile = context.getVirtualDocument(containingFile)
                        .fsPath;
                }
                const containingDir = Path.dirname(containingFile);
                const perFolderCache = cache.getOrCreateCacheForDirectory(containingDir);
                const result = moduleNames.map(moduleName => {
                    var _a;
                    if (vueVirtualTextdocument.isVueFile(moduleName)) {
                        if (perFolderCache.has(moduleName)) {
                            return perFolderCache.get(moduleName).resolvedModule;
                        }
                        const result = ts.resolveModuleName(moduleName, containingFile, options, {
                            ...info.serverHost,
                            fileExists(fileName) {
                                fileName = fileName.endsWith('.vue.ts')
                                    ? fileName.substr(0, fileName.length - 3)
                                    : fileName;
                                return info.serverHost.fileExists(fileName);
                            },
                        }, undefined, redirectedReference);
                        if ((_a = result.resolvedModule) === null || _a === void 0 ? void 0 : _a.resolvedFileName.endsWith('.vue.ts')) {
                            const fileName = result.resolvedModule.resolvedFileName.replace(/\.vue\.ts$/, '.vue');
                            if (info.serverHost.fileExists(fileName)) {
                                const document = context
                                    .getVueDocument(fileName)
                                    .getBlockDocument('script');
                                const resolvedModule = {
                                    isExternalLibraryImport: moduleName.includes('node_modules'),
                                    resolvedFileName: document.fsPath,
                                    extension: fileNameToExtension(document.fsPath),
                                };
                                perFolderCache.set(moduleName, { resolvedModule });
                                return resolvedModule;
                            }
                        }
                    }
                    else {
                        const [result] = fn([moduleName], containingFile, reusedNames, redirectedReference, options);
                        return result;
                    }
                });
                if (info.project.projectService.logger.loggingEnabled()) {
                    info.project.projectService.logger.info(`vue:: ${key}.resolveModuleNames in ${containingFile}`);
                    info.project.projectService.logger.startGroup();
                    moduleNames.forEach((moduleName, index) => {
                        var _a;
                        info.project.projectService.logger.info(`  ${moduleName} => ${(_a = result[index]) === null || _a === void 0 ? void 0 : _a.resolvedFileName}`);
                    });
                    info.project.projectService.logger.endGroup();
                }
                return result;
            });
        }
        function patchScriptInformation(key) {
            // Script Information
            if (info[key].getScriptKind) {
                override(info[key], 'getScriptKind', fn => path => {
                    let kind;
                    if (vueVirtualTextdocument.isVueFile(path)) {
                        kind = ts.ScriptKind.External;
                    }
                    else if (vueVirtualTextdocument.isVirtualFile(path)) {
                        kind = languageIdToScriptKind(context.getVirtualDocument(path).languageId);
                    }
                    else {
                        kind = fn(path);
                    }
                    return kind;
                });
            }
            if (info[key].getScriptVersion) {
                override(info[key], 'getScriptVersion', fn => path => {
                    let version = '0';
                    if (vueVirtualTextdocument.isVueFile(path)) {
                        const doc = context.getVueDocument(path);
                        version = 'Vue-' + doc.version;
                    }
                    else if (vueVirtualTextdocument.isVirtualFile(path)) {
                        const doc = context.getVirtualDocument(path);
                        version = 'Vue-' + doc.container.version + '-' + doc.version;
                    }
                    else {
                        version = fn(version);
                    }
                    return version;
                });
            }
            if (info[key].getScriptSnapshot) {
                override(info[key], 'getScriptSnapshot', fn => path => vueVirtualTextdocument.isVirtualFile(path)
                    ? ts.ScriptSnapshot.fromString(context.getVirtualDocument(path).getText())
                    : fn(path));
            }
        }
        function patchProjectService() {
            override(info.project.projectService, 'getScriptInfoForPath', fn => fileName => {
                const result = fn(fileName);
                if (!result && vueVirtualTextdocument.isVirtualFile(fileName)) {
                    log('getScriptInfoForPath file=' + fileName);
                    const document = context.getVirtualDocument(fileName);
                    const result = new ts.server.ScriptInfo(info.serverHost, ts.server.toNormalizedPath(document.fsPath), languageIdToScriptKind(document.languageId), false, fileName);
                    // @ts-ignore - UNSAFE accessing private property.
                    info.project.projectService.filenameToScriptInfo.set(result.path, result);
                    return result;
                }
                else if (result && vueVirtualTextdocument.isVueFile(fileName)) {
                    try {
                        patchScriptInfo(fileName, result);
                    }
                    catch (_a) { }
                }
                return result;
            });
        }
        function patchScriptInfo(fileName, result) {
            if (!result)
                return;
            if (!vueVirtualTextdocument.isVueFile(fileName))
                return;
            // @ts-ignore hidden property on patched script info.
            if (!result.__vue__) {
                const uri = vscodeUri.URI.file(fileName).toString();
                const editContent = result.editContent.bind(result);
                Object.defineProperty(result, '__vue__', {
                    enumerable: false,
                    configurable: false,
                    value: true,
                });
                result.editContent = function (start, end, text) {
                    editContent(start, end, text);
                    const version = getLastNumberFromVersion(result.getLatestVersion());
                    const document = context.getVueDocument(uri);
                    const snapshot = result.getSnapshot();
                    vueVirtualTextdocument.VueTextDocument.update(document, [{ text: snapshot.getText(0, snapshot.getLength()) }], version);
                };
                const document = context.getVueDocument(uri);
                const script = document.getBlockDocument('script');
                if (script) {
                    info.project.projectService.openClientFileWithNormalizedPath(ts.server.toNormalizedPath(script.fsPath));
                }
            }
        }
        patchFileSystem('serverHost');
        patchScriptInformation('languageServiceHost');
        patchModuleResolution('languageServiceHost');
        patchProjectService();
        return VueLanguageServer.create(context, info.languageService, ts);
    }
    function languageIdToScriptKind(languageId) {
        switch (languageId) {
            case 'typescript':
                return ts.ScriptKind.TS;
            case 'javascript':
                return ts.ScriptKind.JS;
            case 'vue':
                return ts.ScriptKind.External;
        }
        return ts.ScriptKind.Unknown;
    }
    function fileNameToExtension(fileName) {
        const ext = fileName.substr(fileName.lastIndexOf('.'));
        switch (ext) {
            case '.ts':
                return ts.Extension.Ts;
            case '.js':
                return ts.Extension.Js;
            case '.tsx':
                return ts.Extension.Tsx;
            case '.jsx':
                return ts.Extension.Jsx;
            case '.json':
                return ts.Extension.Json;
            default:
                return ts.Extension.Js;
        }
    }
    return { create };
}
function getLastNumberFromVersion(version) {
    const parts = version.split(/[^0-9]+/);
    const ver = parts.pop();
    return Number(ver);
}
function override(object, key, handler) {
    // @ts-ignore
    const original = object[key].bind(object);
    const fn = handler(original);
    // @ts-ignore
    object[key] = fn;
}

module.exports = init;
//# sourceMappingURL=typescript-plugin-vue.js.map
